{% import 'keg_elements/forms/horizontal_b4.html' as horizontal %}
{% extends 'keg_auth/form-base.html' %}

{% block form %}
{{ horizontal.form(form, action_text=action_button_text, cancel_url=cancel_url) }}
{% endblock %}
