{% extends config.get('BASE_TEMPLATE') or config['KEGAUTH_BASE_TEMPLATE'] %}

{% block title %}{{ page_title }} | {{ super() }}{% endblock %}

{% block page_content %}
    {# Show any flash messages. #}
{% endblock %}
