{% extends "base-page.html" %}

{% block page_content %}
    <div class="jumbotron">
        <h1>Keg Auth Demo</h1>
    </div>
    {% if current_user.is_authenticated %}
        <p>Hi {{current_user.display_value}}!</p>
        <p><a href="{{auth_manager.url_for('logout')}}">Logout.</a>
    {% else %}
        <p>You probably want to head to the <a href="{{auth_manager.url_for('login')}}">login page.</a></p>
    {% endif %}
{% endblock %}
