{% import 'keg_elements/forms/horizontal.html' as horizontal %}
{% extends config.get('BASE_TEMPLATE') or config['KEGAUTH_BASE_TEMPLATE'] %}

{% block title %}{{ page_title }}{% endblock %}

{% block scripts %}
{{ super() }}
{% if (use_select2 == true) %}
    {% include 'keg_auth/select2-scripts.html' %}
{% endif %}
{{ horizontal.custom_js() }}
{% endblock %}

{% block styles %}
{{ super() }}
{% if (use_select2 == true) %}
    {% include 'keg_auth/select2-styles.html' %}
{% endif %}
{{ horizontal.custom_css() }}
{% endblock %}

{% block page_content %}
<h1>{{ page_heading }}</h1>
{% block form %}{% endblock %}
{% endblock %}
