{% import "bootstrap/utils.html" as utils %}
{% import "keg_auth/navigation.html" as navigation with context %}
{% extends "base.html" %}

{% block content %}
<div id="navigation">
    {% if auto_expand_menu %}
    {{ navigation.render_menu(auth_manager.menus['main'], expand_to_current=True) }}
    {% else %}
    {{ navigation.render_menu(auth_manager.menus['main']) }}
    {% endif %}
</div>
<div class="container" id="flash-messages">
    {%- with messages = get_flashed_messages(with_categories=True) %}
    {%- if messages %}
      <div class="row">
        <div class="col-md-12">
          {{utils.flashed_messages(messages)}}
        </div>
      </div>
    {%- endif %}
    {%- endwith %}
</div>
<div class="container" id="page-content">
    {% block page_content %}{% endblock %}
</div>
{% endblock %}
